<?php
/**
 * MyResearch Controller
 *
 * Copyright (C) 2019 Leipzig University Library
 *
 * PHP version 7
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * @category VuFind
 * @package  Controller
 * @author   Robert Lange <lange@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
namespace fid\Controller;

use fid\Service\Client;
use fid\Service\ClientException;
use fid\Service\DataTransferObject\Library;
use fid\Service\DataTransferObject\User;
use VuFind\Exception\Forbidden;
use VuFind\Exception\ListPermission as ListPermissionException;
use VuFind\Search\RecommendListener;
use Zend\ServiceManager\ServiceLocatorInterface;

/**
 * Controller for the user account area.
 *
 * @category VuFind
 * @package  Controller
 * @author   Robert Lange <lange@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
class MyResearchController extends \VuFind\Controller\MyResearchController
{
    /**
     * Main Configuration
     *
     * @var \Zend\Config\Config
     */
    protected $mainConfig;

    /**
     * Fidis client
     *
     * @var \fid\Service\Client
     */
    protected $fidClient;

    /**
     * MyResearchController constructor.
     *
     * @param ServiceLocatorInterface $serviceLocator Container interfafce
     */
    public function __construct(
        ServiceLocatorInterface $serviceLocator
    ) {
        parent::__construct($serviceLocator);
        $this->mainConfig = $this->getConfig();
    }

    /**
     * Set fidis client
     *
     * @param Client $fidClient fidis client
     *
     * @return void
     */
    public function setFidClient(Client $fidClient): void
    {
        $this->fidClient = $fidClient;
    }

    /**
     * Show links for Access to Specific and Press Databases, E-Books
     *
     * @return mixed
     * @throws \VuFind\Exception\ILS
     */
    public function databasesAction()
    {
        if (($user = $this->getUser())) {
            $patron = $this->catalogLogin();
        }

        $hasDbis        = $this->hasDbis();
        $hasLicenses    = $this->hasLicenses();

        if (!$hasDbis && !$hasLicenses) {
            return $this->forwardTo('MyResearch', 'Profile');
        }

        $view = $this->createViewModel(['user' => $user ?? null]);

        if ($hasDbis) {
            $view->isDbisAvailable = true;

            if ($this->fidClient->isLoggedOn() && isset($patron['user'])) {
                try {
                    $library = $this->getHomeLibrary($patron);
                    if (null !== $library) {
                        $view->homeLibrary = $library->getDbis();
                    }
                    $view->map_bibid_to_dbis = false;
                } catch (ClientException $exception) {
                    // TODO: Implement logging (in database?)
                    $view->homeLibrary = $patron['user']->getHomeLibrary();
                    $view->map_bibid_to_dbis = true;
                }
            }
        }

        if ($hasLicenses) {
            $view->licenses = $this->getLicenses();
        }

        if (!empty($patron)) {
            $catalog = $this->getILS();
            $this->addAccountBlocksToFlashMessenger($catalog, $patron);
        }

        $view->setTemplate('myresearch/databases');
        return $view;
    }

    /**
     * Check if DBIS module is active and user has permission
     *
     * @return bool
     */
    public function hasDbis(): bool
    {
        // Is module activated?
        if (!isset($this->mainConfig['ExternalDatabases']['DBIS'])
            || $this->mainConfig['ExternalDatabases']['DBIS'] != true
        ) {
            return false;
        }

        // Has user rights to access?
        try {
            return $this->permission()->check('access.DBIS', 'exception') == null;
        } catch (Forbidden $e) {
            // Permission Plugin needs cumbersome exception for logged-in users
        }

        return false;
    }

    /**
     * Get patrons home libray if saved by patron
     *
     * @param $patron patron
     *
     * @return \fid\Service\DataTransferObject\Library | null
     * @throws \fid\Service\ClientException
     */
    protected function getHomeLibrary($patron): ?Library
    {
        if (is_a($patron['user'], 'fid\Service\DataTransferObject\User')
            && $homeLibrary = $patron['user']->getHomeLibrary()
        ) {
            if (isset($patron['libs'][$homeLibrary])) {
                return $patron['libs'][$homeLibrary];
            }

            /* no home library found in patron object?
             * then fetch library from api directly */
            return $this->fidClient->requestLibraryById($homeLibrary);
        }

        return null;
    }

    /**
     * Check if licenses are specified and user has permission
     *
     * @return bool
     */
    public function hasLicenses(): bool
    {
        // Are licenses configured
        if (!isset($this->mainConfig['MediaLicenses'])
            || count($this->mainConfig['MediaLicenses']) == 0
        ) {
            return false;
        }

        // Has user rights to access?
        try {
            return
                $this->permission()->check('access.Licenses', 'exception') == null;
        } catch (Forbidden $e) {
            // Permission Plugin needs cumbersome exception for logged-in users
        }

        return false;
    }

    /**
     * Get links for licensed resources
     *
     * @return mixed
     */
    protected function getLicenses()
    {
        $licenses = false;
        $configLicenses = $this->mainConfig['MediaLicenses'] ?? [];

        if (count($configLicenses) > 0) {
            $licenses = [];
            $i = 0;
            foreach ($configLicenses as $key => $value) {
                $licenses[$i]['name'] = $key;
                $licenses[$i]['url'] = $value;
                $licenses[$i]['desc']
                    = $this->translate('licenses_' . $key . '_desc', [], '');
                $i++;
            }
        }

        return $licenses;
    }

    /**
     * Send user's saved favorites from a particular list to the view
     *
     * @return mixed
     */
    public function mylistAction()
    {
        // Fail if lists are disabled:
        if (!$this->listsEnabled()) {
            throw new ForbiddenException('Lists disabled');
        }

        // Check for "delete item" request; parameter may be in GET or POST depending
        // on calling context.
        $deleteId = $this->params()->fromPost(
            'delete',
            $this->params()->fromQuery('delete')
        );
        if ($deleteId) {
            $deleteSource = $this->params()->fromPost(
                'source',
                $this->params()->fromQuery('source', DEFAULT_SEARCH_BACKEND)
            );
            // If the user already confirmed the operation, perform the delete now;
            // otherwise prompt for confirmation:
            $confirm = $this->params()->fromPost(
                'confirm',
                $this->params()->fromQuery('confirm')
            );
            if ($confirm) {
                $layout = $this->params()->fromPost(
                    'layout',
                    $this->params()->fromQuery('layout')
                );

                /* #17712 not necessary to fetch items after deleting by ajax */
                $success = $this->performDeleteFavorite($deleteId, $deleteSource);
                if ($layout === 'lightbox' || $success !== true) {
                    return $success;
                }
            } else {
                return $this->confirmDeleteFavorite($deleteId, $deleteSource);
            }
        }

        // If we got this far, we just need to display the favorites:
        try {
            $runner = $this->serviceLocator->get('VuFind\Search\SearchRunner');

            // We want to merge together GET, POST and route parameters to
            // initialize our search object:
            $request = $this->getRequest()->getQuery()->toArray()
                + $this->getRequest()->getPost()->toArray()
                + ['id' => $this->params()->fromRoute('id')];

            // Set up listener for recommendations:
            $rManager = $this->serviceLocator
                ->get('VuFind\Recommend\PluginManager');
            $setupCallback = function ($runner, $params, $searchId) use ($rManager) {
                $listener = new RecommendListener($rManager, $searchId);
                $listener->setConfig(
                    $params->getOptions()->getRecommendationSettings()
                );
                $listener->attach($runner->getEventManager()->getSharedManager());
            };

            $results = $runner->run($request, 'Favorites', $setupCallback);
            return $this->createViewModel(
                [
                    'params' => $results->getParams(),
                    'results' => $results,
                    'context' => "favorite"
                ]
            );
        } catch (ListPermissionException $e) {
            if (!$this->getUser()) {
                return $this->forceLogin();
            }
            throw $e;
        }
    }
}
