<?php
/**
 * Fidis client config listener
 *
 * Copyright (C) 2020 Leipzig University Library
 *
 * PHP Version 7
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * @category VuFind
 * @package  Listener
 * @author   Sebastian Kehr <kehr@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU GPLv2
 * @link     https://vufind.org/wiki/development Wiki
 */
namespace fid\Listener;

use fid\Service\ClientException;
use Psr\Container\ContainerInterface;
use VuFind\Config\PluginManager;
use VuFind\Cookie\CookieManager;
use Zend\Config\Config;
use Zend\EventManager\AbstractListenerAggregate;
use Zend\EventManager\EventManagerInterface;
use Zend\Mvc\MvcEvent;
use Zend\View\Model\ViewModel;

/**
 * Fidis client config listener
 *
 * @category VuFind
 * @package  Listener
 * @author   Sebastian Kehr <kehr@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU GPLv2
 * @link     https://vufind.org/wiki/development Wiki
 */
class ClientConfigListener extends AbstractListenerAggregate
{
    protected const PARAM = 'fidis_base_url';

    protected const EVENT = MvcEvent::EVENT_ROUTE;

    protected const CONFIG_KEY = 'fid';

    protected const ERROR_CONFIG_FILE = 'Invalid configuration file "%s.ini".';

    protected const STATUS = '%sFidis: %s';

    /* @var array */
    protected $config;

    /**
     * Get configuration
     *
     * @return array
     */
    public function getConfig(): array
    {
        return $this->config;
    }

    /**
     * Attach onDispatch method to event (MvcEvent::EVENT_ROUTE)
     *
     * @param EventManagerInterface $events   Event manager
     * @param $priority Used priority
     *
     * @return void
     */
    public function attach(EventManagerInterface $events, $priority = 9000)
    {
        $events->attach(static::EVENT, [$this, 'onDispatch'], $priority);
    }

    /**
     * Handling fidis base url update via GET parameter (if allowed in config)
     *
     * @param MvcEvent $event Event
     *
     * @return void
     * @throws ClientException
     */
    public function onDispatch(MvcEvent $event)
    {
        /* @var Config $config */
        /* @var PluginManager $configManager */
        /* @var ContainerInterface $container */
        $request = $event->getRequest();
        $container = $event->getApplication()->getServiceManager();
        $configManager = $container->get(PluginManager::class);
        $config = $configManager->get(self::CONFIG_KEY)->Client;

        if (!$config instanceof Config || !$config->baseUrl) {
            $message = sprintf(self::ERROR_CONFIG_FILE, self::CONFIG_KEY);
            throw new ClientException($message);
        }

        $this->config = $config->toArray();
        if (!$config->baseUrlOverrideAllowed) {
            return;
        }

        $baseUrl = $request->getQuery(self::PARAM)
            ?? $request->getCookie()[self::PARAM]
            ?? $config->baseUrl;

        $valid = $baseUrl === $config->baseUrl
            || preg_match("#$config->baseUrlOverridePattern#", $baseUrl);

        if (!$valid) {
            $baseUrl = $config->baseUrl;
        }

        /* @var ViewModel $viewModel */
        /* @var CookieManager $cookieManager */
        $this->config['baseUrl'] = $baseUrl;
        $cookieManager = $container->get(CookieManager::class);
        $cookieManager->set(self::PARAM, $baseUrl);
        $viewModel = $container->get('ViewManager')->getViewModel();
        $fidisStatus = sprintf(self::STATUS, $valid ? '' : '!', $baseUrl);
        $viewModel->setVariable('fidisStatus', $fidisStatus);
    }
}
