<?php
/**
 * Dto for fid order
 *
 * Copyright (C) 2019 Leipzig University Library
 *
 * PHP version 7
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * @category VuFind
 * @package  Service
 * @author   Sebastian Kehr <kehr@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
namespace fid\Service\DataTransferObject;

namespace fid\Service\DataTransferObject;

use Symfony\Component\Serializer\Annotation\Groups;

/**
 * FID order dto class
 *
 * @category VuFind
 * @package  Service
 * @author   Sebastian Kehr <kehr@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
class Order
{
    /**
     * Identifier
     *
     * @var      string|null
     * @Groups({
     *     "order:creation:response",
     *     "order:list:response",
     *     "order:update:request",
     *     "user:details:response",
     *     "user:creation:response",
     *     "user:update:response"
     * })
     */
    protected $id;

    /**
     * Type of order
     *
     * @var      string
     * @Groups({
     *     "order:creation:request",
     *     "order:creation:response",
     *     "order:update:request",
     *     "order:list:response",
     *     "user:details:response",
     *     "user:creation:response",
     *     "user:update:response"
     * })
     */
    protected $type = '';

    /**
     * Label
     *
     * @var      string
     * @Groups({
     *     "order:creation:request",
     *     "order:creation:response",
     *     "order:list:response",
     *     "order:update:request",
     *     "user:details:response",
     *     "user:creation:response",
     *     "user:update:response"
     * })
     */
    protected $label = '';

    /**
     * User
     *
     * @var      User|null
     * @Groups({
     *     "order:creation:request",
     *     "order:creation:response",
     *     "order:list:response",
     *     "order:update:request",
     *     "user:details:response",
     *     "user:creation:response",
     *     "user:update:response"
     * })
     */
    protected $user;

    /**
     * Custom data
     *
     * @var      array
     * @Groups({
     *     "order:creation:request",
     *     "order:creation:response",
     *     "order:update:request",
     *     "order:list:response",
     *     "user:details:response",
     *     "user:creation:response",
     *     "user:update:response"
     * })
     */
    protected $data = [];

    /**
     * Creation time
     *
     * @var      \DateTime
     * @Groups({
     *     "order:creation:response",
     *     "order:list:response",
     *     "order:update:request",
     *     "user:details:response",
     *     "user:creation:response",
     *     "user:update:response"
     * })
     */
    protected $createdAt;

    public const DATETIME_FORMAT = 'Y-m-d H:i:s';

    /**
     * Get identifier
     *
     * @return string|null
     */
    public function getId(): ?string
    {
        return $this->id;
    }

    /**
     * Set identifier
     *
     * @param string|null $id identifier
     *
     * @return void
     */
    public function setId(?string $id): void
    {
        $this->id = $id;
    }

    /**
     * Get label
     *
     * @return string
     */
    public function getLabel(): string
    {
        return $this->label;
    }

    /**
     * Set label
     *
     * @param string $label label
     *
     * @return void
     */
    public function setLabel(string $label): void
    {
        $this->label = $label;
    }

    /**
     * Get type of order
     *
     * @return string
     */
    public function getType(): string
    {
        return $this->type;
    }

    /**
     * Set type of order
     *
     * @param string $type type
     *
     * @return void
     */
    public function setType(string $type): void
    {
        $this->type = $type;
    }

    /**
     * Get fid user
     *
     * @return User|null
     */
    public function getUser(): ?User
    {
        return $this->user;
    }

    /**
     * Set fid user
     *
     * @param User|null $user user
     *
     * @return void
     */
    public function setUser(?User $user): void
    {
        $this->user = $user;
    }

    /**
     * Get data - container for additional miscellaneous information
     *
     * @return array
     */
    public function getData(): array
    {
        return $this->data;
    }

    /**
     * Set data - container for additional miscellaneous information
     *
     * @param array $data custom data
     *
     * @return void
     */
    public function setData(array $data): void
    {
        $this->data = $data;
    }

    /**
     * Get creation date
     *
     * @return \DateTime
     */
    public function getCreatedAt(): \DateTime
    {
        return $this->createdAt;
    }

    /**
     * Set creation date
     *
     * @param \DateTime $createdAt created at
     *
     * @return void
     */
    public function setCreatedAt(\DateTime $createdAt): void
    {
        $this->createdAt = $createdAt;
    }

    /**
     * Get order label
     *
     * @return string
     */
    public function __toString()
    {
        return $this->getLabel();
    }

    /**
     * Get due date
     *
     * @return \DateTime|null
     */
    public function getDueDate(): ? \DateTime
    {
        return (!empty($this->data['dueDate']))
            ? \DateTime::createFromFormat(
                self::DATETIME_FORMAT,
                $this->data['dueDate']
            )
            : null;
    }

    /**
     * Set due date
     *
     * @param \DateTime $date due date
     *
     * @return void
     */
    public function setDueDate(\DateTime $date)
    {
        $data = $this->getData();
        $data['dueDate'] = $date->format(self::DATETIME_FORMAT);
        $this->setData($data);
    }

    /**
     * Get due date as string d.m.Y.
     *
     * @return String
     */
    public function displayDueDate(): String
    {
        return $this->getDueDate() ? $this->getDueDate()->format('d.m.Y') : '';
    }

    /**
     * Check due date
     *
     * @return bool
     */
    public function isOverdue() : bool
    {
        if ($duedate = $this->getDueDate()) {
            $now = new \DateTime();
            return $now > $duedate ? true : false;
        } else {
            return false;
        }
    }

    /**
     * Get order status
     *
     * @return string
     */
    public function getStatus(): ? string
    {
        return $this->data['status'] ?? null;
    }

    /**
     * Set order status
     *
     * @param string $status status
     *
     * @return void
     */
    public function setStatus(?string $status): void
    {
        if ($status != null) {
            $this->data['status'] = $status;
        }
    }

    /**
     * Get record identifier
     *
     * @return mixed|null
     */
    public function getRecordId()
    {
        return $this->data['record']['id'] ?? null;
    }
}
