<?php
/**
 * Get It box view helper
 *
 * PHP version 7
 *
 * Copyright (C) Leipzig University Library 2019.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Gregor Gawol <gawol@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
namespace fid\View\Helper\Root;

use VuFind\I18n\Translator\TranslatorAwareTrait;
use VuFind\RecordDriver\AbstractBase;
use VuFind\Role\PermissionManager;
use VuFind\View\Helper\Root\Config;
use Zend\View\Helper\AbstractHelper;

/**
 * Get It box view helper
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Gregor Gawol <gawol@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development Wiki
 */
class GetIt extends AbstractHelper
{
    use TranslatorAwareTrait;

    /* @var \Zend\Config\Config */
    protected $config;

    /* @var AbstractBase */
    protected $driver;

    /* @var array list of available source ID definitions from config */
    protected $sids = [
        'source_V1_openaccess',
        'source_V3_magazinarticles',
        'source_V4_pda',
        'source_V5_ebl',
        'source_V6_izi',
        'source_V7_margeretlib',
        'source_V8_vkfilmcollection',
        'source_V9_vkfilm2',
        'source_V10_mediaviewweb',
        'source_V11_base',
    ];

    /**
     * FID API Client
     *
     * @var \fid\Service\Client
     */
    protected $fidClient;

    /**
     * Permission Manager
     *
     * @var PermissionManager
     */
    protected $permissionManager;

    /**
     * Solr field source_id
     *
     * @var
     */
    protected $sid;

    /**
     * Solr field format
     * or instance specific configuration
     *
     * @var
     */
    protected $format;

    /**
     * Solr field facet_avail
     *
     * @var
     */
    protected $facetAvail;

    /**
     * Solr field multipart_set
     *
     * @var
     */
    protected $multipart;

    /**
     * Has a hierarchy parent title
     *
     * @var
     */
    protected $isCollection;

    /**
     * Solr field institution
     *
     * @var
     */
    protected $institution;

    /**
     * Solr field mega_collection
     *
     * @var
     */
    protected $megacollection;

    /**
     * Solr field collection
     *
     * @var
     */
    protected $collection;

    /**
     * Solr field isbn
     *
     * @var
     */
    protected $isbn;

    /**
     * Solr field issn
     *
     * @var
     */
    protected $issn;

    /**
     * Default color of a accordeon
     *
     * @var
     */
    protected $accordeonColorDefault;

    /**
     * Second color of the accordeon
     *
     * @var
     */
    protected $accordeonColorAlternative;

    /**
     * Default headline text of the get it box
     *
     * @var
     */
    protected $accordeonHeadlineDefault;

    /**
     * Second headline text of the get it box
     *
     * @var
     */
    protected $accordeonHeadlineAlternative;

    /**
     * Rules to be evaluated
     *
     * @var Config rules config
     */
    protected $getItRules;

    /**
     * GetIt constructor.
     *
     * @param $getItConfig       Config config
     * @param $fidClient         FIDIS Client
     * @param $permissionManager Permission Manager
     * @param $getItRules        Config Manager
     */
    public function __construct($getItConfig, $fidClient, $permissionManager, $getItRules)
    {
        $this->config = $getItConfig;
        $this->fidClient = $fidClient;
        $this->permissionManager = $permissionManager;

        if (!empty($getItRules)) {
            if (isset($getItRules['rules']) && isset($getItRules['stopFlags'])) {
                $this->getItRules = $getItRules;
            }
        }
    }

    /**
     * Get Solr field data
     *
     * @param $driver Record Driver
     *
     * @return GetIt GetIt View Helper
     */
    public function __invoke($driver)
    {
        $this->driver = $driver;
        $this->sid = $this->driver->tryMethod('getSourceID');
        $this->format = $this->driver->tryMethod('getFormats');
        $this->facetAvail = $this->driver->tryMethod('getFacetAvail');
        //$this->multipart = $this->driver->tryMethod('isMultiPartSet');
        $this->isCollection = $this->driver->tryMethod('hasChildren');
        $this->institution = $this->driver->tryMethod('getInstitutions');
        $this->megacollection = $this->driver->tryMethod('getMegaCollection');
        $this->collection = $this->driver->tryMethod('getCollection');
        $this->isbn = $this->driver->tryMethod('getISBNs');
        $this->issn = $this->driver->tryMethod('getISSNs');

        $this->accordeonColorDefault = 'azure';
        $this->accordeonColorAlternative = 'amber';
        $this->accordeonHeadlineDefault = $this->translate('getit_infos');
        $this->accordeonHeadlineAlternative = $this->translate('getit_links');

        return $this;
    }

    /**
     * Check if User is authorized
     *
     * @param $permission Permission
     *
     * @return bool
     */
    public function hasPermission($permission)
    {
        return $this->permissionManager->isAuthorized($permission);
    }

    /**
     * Retrieve SIDs from config (getit.ini)
     *
     * @param $value sid
     *
     * @return array
     */
    protected function _getSourceIds($value)
    {
        $sids = $this->config->SourceIds;
        return isset($sids->$value) ? explode(',', $sids->$value) : [];
    }

    /**
     * Retrieve NonSIDs from config (getit.ini)
     *
     * @param $value sid
     *
     * @return string
     */
    protected function _getNonSourceIds($value)
    {
        $sids = $this->config->NonSourceIds;
        return $sids->$value ?? '';
    }

    /**
     * Read specific config of given record
     * Set default Values of getit box
     *
     * @return array config of given record
     */
    public function getConfig()
    {
        //set default values
        $accordeonColor = $this->accordeonColorDefault;
        $accordeonHeadline = $this->accordeonHeadlineDefault;
        $boxHeadline = $this->translate('Get it');
        $notice = $this->translate('getit_text_default');
        $showLinks = true;
        $showOrderButton = false;
        $showPartCopyButton = false;
        $showArticleButton = false;
        $bossData = false;
        $isEBCEBooks = false;
        $hideNotice = false;
        $isAiSidRecord = false;
        $noticeLinkType = 'register';

        // let specific functions override defaults where necessary
        foreach ($this->sids as $sid_config) {
            $sid = $this->_getSourceIds($sid_config);
            if (current($sid) === 'ALL' || in_array($this->sid, $sid)) {
                $dynamic_func = 'get_config_' . $sid_config;
                $this->$dynamic_func(
                    $accordeonColor,
                    $accordeonHeadline,
                    $boxHeadline,
                    $notice,
                    $showLinks,
                    $showOrderButton,
                    $showPartCopyButton,
                    $showArticleButton,
                    $bossData,
                    $isEBCEBooks,
                    $hideNotice,
                    $isAiSidRecord,
                    $noticeLinkType
                );
                if ($this->sid !== '0') {
                    //special case for source ID 0
                    // this is th only SID that is contained in multiple configs (V8 + V9)
                    break;
                }
            }
        }
        # injecting rules evaluator for testing
        if (isset($this->getItRules)) {
            $eval = new \finc\Rules\Evaluator\Evaluator(
                array_values($this->getItRules['rules']),
                (array)$this->getItRules['stopFlags']
            );
            $rvh = $this->getView()->plugin('record');
            $context = [
                'this' => $this,
                'client' => $this->fidClient,
                'permissionmanager' => $this->permissionManager,
                'driver' => $this->driver,
                'allowDirectDelivery' => $rvh($this->driver)->allowDirectDelivery(),
                'sid' => $this->sid,
                'format' => $this->format,
                'facet_avail' => $this->facetAvail,
                'is_collection' => $this->isCollection,
                'institution' => $this->institution,
                'megacollection' => $this->megacollection,
                'collection' => $this->collection,
                'isbn' => $this->isbn,
                'issn' => $this->issn,
                'allowDigitizationOnDemand' => $rvh($this->driver)->allowDigitizationOnDemand(),
                'enabledServices' => $this->config->Permission
            ];
            // set variables of config in context
            foreach ($this->getItRules as $key => $value) {
                if ($key != "rules" and $key != "stopFlags") {
                    $context[$key] = $value;
                }
            }
            $context = $eval($context);

            /* tmp checks - TODO: Remove*/
            if ('development' == APPLICATION_ENV || 'testing' == APPLICATION_ENV) {
                $checkProps = [];
                // Part 1
                array_push($checkProps, 'accordeonColor', 'accordeonHeadline', 'boxHeadline', 'notice');
                // Part 2
                array_push($checkProps, "isAiSidRecord", "hideNotice", "isEBCEBooks", "noticeLinkType", "bossData");
                // Part 3
                array_push($checkProps, "showLinks", "showOrderButton", "showPartCopyButton", "showArticleButton");

                // just for easier debugging - delete later
                $debugComparer = [];

                foreach ($checkProps as $prop) {
                    if (${$prop} === $context["{$prop}"]) {
                        $debugComparer[] = $prop . " is identical: " . $context["{$prop}"];
                    } else {
                        $debugComparer[] = $prop . " is '" . $context["{$prop}"] . "', but should be instead: '" . ${$prop} . "'";
                    }

                    assert(
                        ${$prop} == $context["{$prop}"],
                        new \Exception($this->driver->getUniqueID() . ":" . $prop . " is '" . $context["{$prop}"]
                        . "', but should be instead: '" . ${$prop} . "'")
                    );
                }
            }
            /* tmp checks - END */
        }

        return $context;
    }

    /**
     * Helper function to check simple intersect
     *
     * @param array $a first array
     * @param array $b second array
     *
     * @return bool
     */
    public function isArrayIntersect($a, $b)
    {
        return !empty(array_intersect($a, $b));
    }

    /**
     * Helper function to promt notices in short syntax
     *
     * @param string $key translation key suffix after "getit_text_"
     *
     * @return string
     */
    public function out($key)
    {
        return $this->translate("getit_text_{$key}");
    }

    /**
     * Open Access
     *
     * Configuration of SIDs 13,22,26,28,30,34,53,80,84,87,88,99,101,150,153,170
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V1_openaccess(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        $accordeonColor = $this->accordeonColorAlternative;
        $accordeonHeadline = $this->accordeonHeadlineAlternative;
        $notice = $this->translate('getit_text_free_avail');
    }

    /**
     * Magazine Articles
     *
     * Configuration of SID 49,55,68,105
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V3_magazinarticles(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        $accordeonColor = $this->accordeonColorAlternative;
        $accordeonHeadline = $this->accordeonHeadlineAlternative;
        $boxHeadline = $this->fidClient->isLoggedOn() ?
            $this->translate('Get it') : $this->translate('getit_logged_not');
        $notice = '';
        $isAiSidRecord = true;
        $showArticleButton = $this->config->Permission->allowSubitoArticleCopy ?? true;
    }

    /**
     * PDA
     *
     * Configuration of SID 51
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V4_pda(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        $boxHeadline = $this->fidClient->isLoggedOn() ?
            $this->translate('Get it') : $this->translate('getit_logged_not');
        if ($this->fidClient->isLoggedOn()) {
            $notice = $this->translate('getit_text_direct_delivery');
        } else {
            $notice = $this->translate('getit_text_free_loan');
            $noticeLinkType = "register";
        }
        $showOrderButton = $this->config->Permission->allowPDA ?? true;
    }

    /**
     * EBL
     *
     * Configuration of SID 76
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V5_ebl(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        $hasEBookAccess = false;
        if ($this->fidClient->isLoggedOn()) {
            $hasEBookAccess = $this->hasPermission('fid.EBooks');
        }

        $accordeonColor = $this->accordeonColorAlternative;
        $accordeonHeadline = $this->accordeonHeadlineAlternative;
        if ($this->fidClient->isLoggedOn()) {
            if ($hasEBookAccess) {
                $notice = $this->translate('getit_text_fidbbi_licence');
            } else {
                $notice = $this->translate('getit_text_13b_1');
                $noticeLinkType = "contact";
            }
        } else {
            $notice = $this->translate('getit_text_info_req_registry');
            $noticeLinkType = "register";
        }
        $boxHeadline = $this->fidClient->isLoggedOn() ?
            $this->translate('Get it') : $this->translate('getit_logged_not');
        $showLinks = $this->fidClient->isLoggedOn() && $hasEBookAccess;
    }

    /**
     * IZI Database
     *
     * Configuration of SID 78
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V6_izi(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        $notice = $this->translate('getit_text_part_not_free');
    }

    /**
     * Margaret Herrick Library
     *
     * Configuration of SID 103
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V7_margeretlib(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        $accordeonColor = $this->accordeonColorAlternative;
        $accordeonHeadline = $this->accordeonHeadlineAlternative;
        $notice = $this->translate('getit_text_nationallicence');
    }

    /**
     * VK Film collection
     *
     * Configuration of SID 0,109,117,119,127,142,148,151,155
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V8_vkfilmcollection(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        //accordeon
        if (in_array('Free', $this->facetAvail)) {
            $accordeonHeadline = $this->accordeonHeadlineAlternative;
            $accordeonColor = $this->accordeonColorAlternative;
        } elseif (in_array('Online', $this->facetAvail)
            && (preg_grep($this->_getNonSourceIds('getit_V8_media_5'), $this->format)
            || preg_grep($this->_getNonSourceIds('getit_V8_media_6'), $this->format)
            || preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format))
        ) {
            $accordeonHeadline = $this->accordeonHeadlineAlternative;
            $accordeonColor = $this->accordeonColorAlternative;
        }
        //END accordeon
        //box headline
        $isLogin = false;
        if (in_array($this->sid, $this->_getSourceIds('source_V9_vkfilm2'))
            && preg_grep($this->_getNonSourceIds('getit_V8_collection_2'), $this->collection)
        ) {# TODO: find out if ex V8_8 is supposed to be an institution or a format
            $isLogin = true;
        } elseif (in_array('Local', $this->facetAvail)
            && ((preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format)
            && !in_array($this->_getNonSourceIds('getit_V8_institute_1'), $this->institution))
            || preg_grep($this->_getNonSourceIds('getit_V8_media_3'), $this->format))
        ) {
            $isLogin = true;
        } elseif (in_array('Online', $this->facetAvail)
            && !in_array('Free', $this->facetAvail)
            && (preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format))
        ) {
            $isLogin = true;
        }
        if ($isLogin && !$this->fidClient->isLoggedOn()) {
            $boxHeadline = $this->translate('getit_logged_not');
        }
        //END boxheadline
        //notice
        if (in_array($this->sid, $this->_getSourceIds('source_V9_vkfilm2'))) {
            if (preg_grep($this->_getNonSourceIds('getit_V8_collection_2'), $this->collection)) {
                if (!$this->fidClient->isLoggedOn()) {
                    $notice = $this->translate('getit_text_info_req_registry');
                } else {
                    $notice = $this->translate('getit_text_fidbbi_licence');
                }
            } elseif ($this->isCollection) {
                $notice = $this->translate('getit_text_multipart_mono', ['%%hierarchy_tree%%' => $this->translate('hierarchy_tree')]);
            } elseif (preg_grep($this->_getNonSourceIds('getit_V8_megacollection_1'), $this->megacollection)
            ) {
                $notice = $this->translate('getit_text_free_avail');
            } elseif (in_array('Local', $this->facetAvail)
                && (preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format)
                || preg_grep($this->_getNonSourceIds('getit_V8_media_3'), $this->format))
            ) {
                if (!$this->fidClient->isLoggedOn()) {
                    $notice = $this->translate('getit_text_info_req_reg');
                    $noticeLinkType = "register";
                } elseif (preg_grep($this->_getNonSourceIds('getit_V8_media_3'), $this->format)) {
                    $notice = $this->translate('getit_text_media_typ_limitation');
                } else {
                    $notice = '';
                }
            } elseif (in_array('Local', $this->facetAvail)
                && preg_grep($this->_getNonSourceIds('getit_V8_media_2'), $this->format)
            ) {
                $notice = $this->translate('getit_text_no_further_info_notice');
            } elseif (in_array('Free', $this->facetAvail)) {
                $notice = $this->translate('getit_text_free_avail');
            } elseif (in_array('Online', $this->facetAvail)
                && (preg_grep($this->_getNonSourceIds('getit_V8_media_5'), $this->format)
                || preg_grep($this->_getNonSourceIds('getit_V8_media_6'), $this->format))
            ) {
                $notice = $this->translate('getit_text_online_res_money_hint');
            } elseif (in_array('Online', $this->facetAvail)
                && preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format)
            ) {
                if (!$this->fidClient->isLoggedOn()) {
                    $notice = $this->translate('getit_text_info_req_reg');
                    $noticeLinkType = "register";
                } else {
                    $notice = $this->translate('getit_text_access_ebook_hint');
                }
            }
        } elseif ($this->isCollection) {
            $notice = $this->translate('getit_text_multipart_mono', ['%%hierarchy_tree%%' => $this->translate('hierarchy_tree')]);
        } elseif (in_array('Local', $this->facetAvail)
            && preg_grep($this->_getNonSourceIds('getit_V8_media_2'), $this->format)
        ) {
            $notice = $this->translate('getit_text_no_further_info_notice');
        } elseif (in_array('Local', $this->facetAvail)) {
            if (!$this->fidClient->isLoggedOn()) {
                $notice = $this->translate('getit_text_info_req_reg');
                $noticeLinkType = "register";
            } elseif (preg_grep($this->_getNonSourceIds('getit_V8_media_3'), $this->format)) {
                $notice = $this->translate('getit_text_media_typ_limitation');
            } elseif (!preg_grep($this->_getNonSourceIds('getit_V8_media_7'), $this->format)) {
                $notice = '';
            }
        } elseif (in_array('Free', $this->facetAvail)) {
            $notice = $this->translate('getit_text_free_avail');
        } elseif (in_array('Online', $this->facetAvail)
            && (preg_grep($this->_getNonSourceIds('getit_V8_media_5'), $this->format)
            || preg_grep($this->_getNonSourceIds('getit_V8_media_6'), $this->format))
        ) {
            $notice = $this->translate('getit_text_online_res_money_hint');
        } elseif (in_array('Online', $this->facetAvail)
            && preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format)
        ) {
            if (!$this->fidClient->isLoggedOn()) {
                $notice = $this->translate('getit_text_info_req_reg');
                $noticeLinkType = "register";
            } else {
                $notice = $this->translate('getit_text_access_ebook_hint');
            }
        }
        //END notice
        //bossData
        if (!in_array('Free', $this->facetAvail)
            //&& (!empty($this->isbn) || !empty($this->issn))
            && !preg_grep($this->_getNonSourceIds('getit_V8_collection_2'), $this->collection)
        ) {
            if (!in_array('Online', $this->facetAvail)
                && preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format)
            ) {
                if (!$this->isCollection) {
                    $showPartCopyButton = $this->config->Permission->allowSubitoPartialCopy ?? true;
                }
            }
            if (preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format)
            ) {
                $bossData = true;
            } elseif (in_array('Local', $this->facetAvail)
                && preg_grep($this->_getNonSourceIds('getit_V8_media_3'), $this->format)
            ) {
                $bossData = true;
            }
        }
        //END bossData
        //hideNotice
        if ($this->fidClient->isLoggedOn()) {
            if (in_array('Local', $this->facetAvail)) {
                if (preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format)
                    && empty($this->isbn)
                    && !preg_grep($this->_getNonSourceIds('getit_V8_institute_1'), $this->institution)
                    && !preg_grep($this->_getNonSourceIds('getit_V8_collection_2'), $this->collection)
                ) {
                    $hideNotice = true;
                }
                if ($this->isCollection) {
                    $hideNotice = false;
                }
            }
        }
        //END hideNotice
    }

    /**
     * VK Film collection
     *
     * Configuration of SID 0 (special cases)
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V9_vkfilm2(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        //accordeon
        if (preg_grep($this->_getNonSourceIds('getit_V8_collection_2'), $this->collection)
            || preg_grep($this->_getNonSourceIds('getit_V8_megacollection_1'), $this->megacollection)
        ) {
            $accordeonHeadline = $this->accordeonHeadlineAlternative;
            $accordeonColor = $this->accordeonColorAlternative;
        }
        //END accordeon
        //showLinks
        if (preg_grep($this->_getNonSourceIds('getit_V8_collection_2'), $this->collection)
        ) {
            if (!$this->fidClient->isLoggedOn()) {
                $showLinks = false;
            }
        }
        //END showLinks
        //orderButton
        if (in_array('Local', $this->facetAvail)
            && preg_grep($this->_getNonSourceIds('getit_V8_media_1'), $this->format)
            && !preg_grep($this->_getNonSourceIds('getit_V8_institute_1'), $this->institution)
            && !$this->isCollection
            && !preg_grep($this->_getNonSourceIds('getit_V8_collection_2'), $this->collection)
        ) {
            $showOrderButton = $this->config->Permission->allowPDA ?? true;
        }
        //END orderButton
        //EBCEbooks
        if (in_array('Online', $this->facetAvail)
            && preg_grep($this->_getNonSourceIds('getit_V8_collection_2'), $this->collection)
        ) {
            $isEBCEBooks = true;
        }
        //END EBCEbooks
    }

    /**
     * MediathekViewWeb
     *
     * Configuration of SID 169
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V10_mediaviewweb(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        $accordeonColor = $this->accordeonColorAlternative;
        $accordeonHeadline = $this->accordeonHeadlineAlternative;
        $notice = $this->translate('getit_text_digital_access_hint');
    }

    /**
     * Bielefeld Academic Search Engine (BASE)
     *
     * Configuration of SID 126
     *
     * @param $accordeonColor     string  accordeon color
     * @param $accordeonHeadline  string  accordeon headline
     * @param $boxHeadline        string  headline text get it box
     * @param $notice             string  text of box version
     * @param $showLinks          boolean display links
     * @param $showOrderButton    boolean display order button
     * @param $showPartCopyButton boolean display part copy button
     * @param $showArticleButton  boolean display part copy button
     * @param $bossData           boolean display data of BOSS webservice
     * @param $isEBCEBooks        boolean is given record UBL-FID-EBC
     * @param $hideNotice         boolean hide notice
     * @param $isAiSidRecord      boolean is given record an ai record
     * @param $noticeLinkType     string type of link
     *
     * @return void
     */
    public function get_config_source_V11_base(
        &$accordeonColor,
        &$accordeonHeadline,
        &$boxHeadline,
        &$notice,
        &$showLinks,
        &$showOrderButton,
        &$showPartCopyButton,
        &$showArticleButton,
        &$bossData,
        &$isEBCEBooks,
        &$hideNotice,
        &$isAiSidRecord,
        &$noticeLinkType
    ) {
        $notice = $this->translate('getit_text_bielefeld');
    }

    /**
     * Has User ordered certain Record (SID 51)
     *
     * @return bool
     */
    public function isRecordOrdered()
    {
        if ($this->fidClient->isLoggedOn()) {
            $id = $this->driver->tryMethod('getUniqueID');
            $record = $this->fidClient->requestRecord($id);

            return $record->getOrdered();
        }
        return false;
    }
}
