<?php
/**
 * Copyright (C) 2019 Leipzig University Library
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * @author   Robert Lange <lange@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU GPLv2
 */
namespace fid_adlr\Controller;

use fid\Service\Client;
use VuFind\Controller\FeedbackController as BaseFeedbackController;
use Zend\View\Model\ViewModel;

/**
 * Redirects the user to the appropriate default VuFind action.
 *
 * @category VuFind
 * @package  Controller
 * @author   Robert Lange <lange@ub.uni-leipzig.de>
 * @author   Alex Purr <purr@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org Main Site
 */
class FeedbackController extends BaseFeedbackController
{
    /**
     * @var Client
     */
    protected $client;

    /**
     * @param Client $client
     */
    public function setClient(Client $client): void
    {
        $this->client = $client;
    }

    /**
     * Display Feedback form.
     *
     * @return \Zend\View\Model\ViewModel
     */
    public function homeAction()
    {
        return $this->forwardTo('Feedback', 'Preselect');
    }

    /**
     * Handles rendering and submit of preselect form.
     * Form configurations are specified in FeedbackForms.yaml
     *
     * @return \Zend\View\Model\ViewModel
     * @throws \VuFind\Exception\Forbidden
     */
    public function preselectAction() : ViewModel
    {
        $formId = $this->params()->fromRoute('id', $this->params()->fromQuery('id'));
        if (!$formId) {
            $formId = 'PreselectSite';
        }

        $form = $this->serviceLocator->get('VuFind\Form\Form');
        $form->setFormId($formId);

        if (!$form->isEnabled()) {
            throw new \VuFind\Exception\Forbidden("Form '$formId' is disabled");
        }

        $view = $this->createViewModel(compact('form', 'formId'));
        $view->useRecaptcha
            = $this->recaptcha()->active('feedback') && $form->useCaptcha();

        $params = $this->params();
        $form->setData($params->fromPost());

        return $view;
    }

    /**
     * Handles rendering and submit of the real feedback form.
     * Form configurations are specified in FeedbackForms.yaml
     *
     * @return \Zend\View\Model\ViewModel
     * @throws \VuFind\Exception\Forbidden
     */
    public function formAction() : ViewModel
    {
        $formId = $this->params()->fromRoute('id', $this->params()->fromQuery('id'));
        $formWasSubmitted = $formId === "PreselectSite" ? false : $this->formWasSubmitted('submit');

        if (!$formId || $formId == "PreselectSite") {
            $formId = 'FeedbackSite';
        }

        $form = $this->serviceLocator->get('VuFind\Form\Form');
        $form->setFormId($formId);

        if (!$form->isEnabled()) {
            throw new \VuFind\Exception\Forbidden("Form '$formId' is disabled");
        }

        $user = $this->getUser();
        $view = $this->createViewModel(compact('form', 'formId', 'user'));
        $view->useRecaptcha = $this->recaptcha()->active('feedback') && $form->useCaptcha();

        $postParams = $this->params()->fromPost();
        $subjectFromSession = $this->followup()->retrieve("emailSubject");

        if (!isset($postParams["emailSubject"])) {
            $postParams["emailSubject"] = $subjectFromSession ?? "";
        } else {
            $this->followup()->store(["emailSubject" => $postParams["emailSubject"]]);
        }

        $form->setData($postParams);

        // force login if NOT general question or catalogue error
        $emailSubject = $postParams["emailSubject"];
        if (empty($user) && $emailSubject != "general_question" && $emailSubject != "catalogue_error") {
            return $this->forceLogin(null, ["emailSubject" => $emailSubject]);
        }

        if (!$formWasSubmitted) {
            if (!empty($user)) {
                $this->prefillUserInfo($form, $user);
            }
            return $view;
        }

        if (!$form->isValid()) {
            return $view;
        }

        // remove followup session parameter
        $this->followup()->clear("emailSubject");
        [$success, $errorMsg] = $this->sendFeedback($form, $postParams);

        $this->showResponse($view, $form, $success, $errorMsg);

        return $view;
    }

    /**
     * @param $form
     * @param $postParams
     * @return array of success / error result
     */
    protected function sendFeedback($form, $postParams): array
    {
        [$messageParams, $template] = $form->formatEmailMessage($postParams);
        if (!empty($this->getUser()) && $user = $this->client->requestUserDetails()) {
            $messageParams["LiberoID"] = [
                "type" => "text",
                "value" => $user->getLiberoId()
            ];
        }
        $emailMessage = $this->getViewRenderer()->partial($template, ['fields' => $messageParams]);

        [$senderName, $senderEmail] = $this->getSender();

        $replyToName = $postParams["name"];
        $replyToEmail = $postParams["email"];

        $recipients = $form->getRecipient();

        $emailSubject = $form->getEmailSubject($postParams);

        foreach ($recipients as $recipient) {
            [$success, $errorMsg] = $this->sendEmail(
                $recipient['name'],
                $recipient['email'],
                $senderName,
                $senderEmail,
                $replyToName,
                $replyToEmail,
                $emailSubject,
                $emailMessage
            );
            if (!$success) {
                return [$success, $errorMsg];
            }
        }

        return [$success, null];
    }

    /**
     * Get form elements
     *
     * @param array $config Form configuration
     *
     * @return array
     */
    protected function getFormElements($config)
    {
        $elements = [];
        foreach ($config['fields'] as $field) {
            if (!isset($field['type'])) {
                continue;
            }
            $elements[] = $field;
        }
        return $elements;
    }
}
