<?php
/**
 * Factory for record driver data formatting view helper
 *
 * PHP version 7
 *
 * Copyright (C) 2020 Leipzig University Library
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Gregor Gawol <gawol@ub.uni-leipzig.de>
 * @author   Robert Lange <lange@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:architecture:record_data_formatter
 * Wiki
 */
namespace fid_adlr\View\Helper\Root;

use VuFind\View\Helper\Root\RecordDataFormatter;
use VuFind\View\Helper\Root\RecordDataFormatter\SpecBuilder;

/**
 * Factory for record driver data formatting view helper
 *
 * @category VuFind
 * @package  View_Helpers
 * @author   Gregor Gawol <gawol@ub.uni-leipzig.de>
 * @author   Robert Lange <lange@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:architecture:record_data_formatter
 * Wiki
 */
class RecordDataFormatterFactory extends \fid\View\Helper\Root\RecordDataFormatterFactory
{
    /**
     * Create the helper.
     *
     * @return RecordDataFormatter
     */
    public function __invoke()
    {
        $helper = parent::__invoke();

        $helper->setDefaults(
            'toc',
            [$this, 'getTOCSpecs']
        );

        return $helper;
    }

    /**
     * Get default specifications for displaying data in core metadata.
     *
     * @return array
     */
    public function getAiCoreSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setTemplateLine(
            'Authors/Corporations',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ]
            ]
        );
        $spec->setTemplateLine(
            'In',
            'getJTitle',
            'data-jTitle.phtml'
        );
        $spec->setTemplateLine(
            'Published',
            'getPublicationDetails',
            'data-publicationDetails.phtml'
        );
        $spec->setLine(
            'Format',
            'getFormats',
            'RecordHelper',
            ['helperMethod' => 'getFormatList']
        );
        return $spec->getArray();
    }

    /**
     * Get default specifications for displaying data in the description tab.
     *
     * @return array
     */
    public function getDefaultDescriptionSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setLine(
            'Dates of publication',
            'getDateSpan'
        );
        $spec->setLine(
            'Item Description',
            'getGeneralNotes'
        );
        $spec->setLine(
            'Physical Description',
            'getPhysicalDescriptions'
        );
        $spec->setLine(
            'Publication Frequency',
            'getPublicationFrequency'
        );
        $spec->setLine(
            'Playing Time',
            'getPlayingTimes'
        );
        $spec->setLine(
            'Format',
            'getSystemDetails'
        );
        $spec->setLine(
            'Audience',
            'getTargetAudienceNotes'
        );
        $spec->setLine(
            'Awards',
            'getAwards'
        );
        $spec->setLine(
            'Production Credits',
            'getProductionCredits'
        );
        $spec->setLine(
            'Bibliography',
            'getBibliographyNotes'
        );
        $spec->setTemplateLine(
            'ISBN',
            'getISBNs',
            'data-isbn.phtml'
        );
        $spec->setTemplateLine(
            'ISSN',
            'getISSNs',
            'data-issn.phtml'
        );
        $spec->setLine(
            'DOI',
            'getCleanDOI'
        );
        $spec->setLine(
            'EISSN',
            'getEISSNs'
        );
        $spec->setLine(
            'Related Items',
            'getRelationshipNotes'
        );
        $spec->setLine(
            'Access',
            'getAccessRestrictions'
        );
        $spec->setLine(
            'Finding Aid',
            'getFindingAids'
        );
        $spec->setLine(
            'Publication_Place',
            'getHierarchicalPlaceNames'
        );
        $spec->setTemplateLine(
            'Author Notes',
            true,
            'data-authorNotes.phtml'
        );
        $spec->setTemplateLine(
            'Call Number',
            'getLocalSignature',
            'data-localSignature.phtml'
        );
        $spec->setLine(
            'Notes',
            'getAdditionalNotes',
            null,
            [
                'useCache' => true
            ]
        );
        $spec->setLine(
            'Note',
            'getContentNote'
        );
        $spec->setTemplateLine(
            'ISBN',
            'getISBNs',
            'data-isbn.phtml'
        );
        $spec->setTemplateLine(
            'Subjects',
            'getAllSubjectHeadings',
            'data-allSubjectHeadings.phtml'
        );
        $spec->setTemplateLine(
            'Published in',
            'getContainerTitle',
            'data-containerTitle.phtml'
        );
        $spec->setLine(
            'New Title',
            'getNewerTitles',
            'data-linkViaFincId.phtml'
        );
        $spec->setLine(
            'Previous Title',
            'getPreviousTitles',
            'data-linkViaFincId.phtml'
        );
        $spec->setTemplateLine(
            'Title Uniform',
            'getTitleUniform',
            'data-titleUniform.phtml',
            [
                'labelFunction' => [$this,'titleUniformLabel']
            ]
        );
        $spec->setLine(
            'Edition',
            'getEdition',
            null,
            ['prefix' => '<span property="bookEdition">', 'suffix' => '</span>']
        );
        $spec->setLine(
            'Dissertation Note',
            'getDissertationNote',
            null,
            ['separator' => ', ']
        );
        $spec->setTemplateLine(
            'Language',
            'getLanguages',
            'data-transEscCommaSepLang.phtml'
        );
        $spec->setTemplateLine(
            'Set Multipart',
            'getHierarchyParentTitle',
            'data-hierarchyParentTitle.phtml'
        );
        $spec->setTemplateLine(
            'Subjects',
            'getAllSubjectHeadings',
            'data-allSubjectHeadings.phtml'
        );
        $spec->setTemplateLine(
            'child_records',
            'getChildRecordCount',
            'data-childRecords.phtml',
            ['allowZero' => false]
        );
        /* finc: remove 'Online Access' in record-details #13770 - VE */
        /* $spec->setTemplateLine('Online Access', true, 'data-onlineAccess.phtml'); */
        $spec->setTemplateLine(
            'Related Items',
            'getAllRecordLinks',
            'data-allRecordLinks.phtml'
        );
        $spec->setMultiLine(
            'Additionals',
            'getAdditionals',
            [$this,'additionals']
        );
        $spec->setLine(
            'Source',
            'getMegaCollection'
        );
        $spec->setMultiLine(
            'Other Relationship Entry',
            'getOtherRelationshipEntry',
            [$this,'otherRelationShipEntry']
        );
        $spec->setLine(
            'Notes',
            'getAdditionalNotes',
            null,
            [
                'useCache' => true
            ]
        );
        $spec->setTemplateLine('Tags', true, 'data-tags.phtml');
        return $spec->getArray();
    }

    /**
     * Get marc specifications for displaying data in core metadata.
     *
     * @return array
     */
    public function getMarcCoreSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setTemplateLine(
            'Authors/Corporations',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ]
            ]
        );
        $spec->setTemplateLine(
            'Title Statement',
            'getTitleStatement',
            'data-titlestatement.phtml'
        );
        $spec->setTemplateLine(
            'Published',
            'getPublicationDetails',
            'data-publicationDetails.phtml'
        );
        $spec->setTemplateLine(
            'Set Multipart',
            'getHierarchyParentTitle',
            'data-hierarchyParentTitle.phtml'
        );
        $spec->setTemplateLine(
            'Format',
            'getFormats',
            'data-format.phtml'
        );
        return $spec->getArray();
    }

    /**
     * Get marc specifications for displaying table of contents.
     *
     * @return array
     */
    public function getTOCSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();
        $spec->setLine(
            'Summary',
            'getSummary',
            null,
            [
                'prefix' => '<p>',
                'separator' => '</p><p>',
                'suffix' => '</p>'
            ]
        );
        $spec->setLine(
            'Table of Contents',
            'getTOC'
        );
        return $spec->getArray();
    }

    /**
     * Get default specifications for displaying data in core metadata.
     *
     * @return array
     */
    public function getDefaultCoreSpecs()
    {
        $spec = new RecordDataFormatter\SpecBuilder();

        $spec->setTemplateLine(
            'Authors/Corporations',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ]
            ]
        );
        $spec->setLine(
            'Published',
            'getPublicationDetails'
        );
        $spec->setTemplateLine(
            'Set Multipart',
            'getHierarchyParentTitle',
            'data-hierarchyParentTitle.phtml'
        );
        $spec->setTemplateLine(
            'Format',
            'getFormats',
            'data-format.phtml'
        );

        return $spec->getArray();
    }

    /**
     * Get general specifications for displaying data in fid pda form.
     *
     * @return SpecBuilder
     */
    public function getFidAcquisitionSpecBuilder()
    {
        $spec = new SpecBuilder();
        $spec->setLine('Title', 'getTitle');
        $spec->setTemplateLine(
            'Authors/Corporations',
            'getDeduplicatedAuthors',
            'data-authors.phtml',
            [
                'useCache' => true,
                'context' => [
                    'requiredDataFields' => [
                        ['name' => 'role', 'prefix' => 'CreatorRoles::']
                    ]
                ]
            ]
        );
        $spec->setTemplateLine(
            'Published',
            'getPublicationDetails',
            'data-publicationDetails.phtml'
        );
        $spec->setTemplateLine(
            'Price',
            'getPrice',
            'data-price.phtml'
        );
        return $spec;
    }
}
