<?php
/**
 * Finc specific Libero ILS Driver for VuFind, using PAIA, DAIA and LiberoWachtl
 * services.
 *
 * PHP version 5
 *
 * Copyright (C) Leipzig University Library 2015.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @category VuFind2
 * @package  ILS_Drivers
 * @author   André Lahmann <lahmann@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     http://vufind.org/wiki/vufind2:building_an_ils_driver Wiki
 */
namespace finc\ILS\Driver;

use finc\ILS\InitException;
use VuFind\Exception\ILS as ILSException;
use VuFind\I18n\Translator\TranslatorAwareInterface;
use VuFind\I18n\Translator\TranslatorAwareTrait;
use VuFindSearch\Query\Query;

/**
 * Finc specific Libero ILS Driver for VuFind, using PAIA, DAIA and LiberoWachtl
 * services.
 *
 * @category VuFind2
 * @package  ILS_Drivers
 * @author   André Lahmann <lahmann@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     http://vufind.org/wiki/vufind2:building_an_ils_driver Wiki
 */
class FincLibero extends FincILS implements TranslatorAwareInterface
{
    public const DELETE_NOTIFICATIONS_SUCCESS = '1';
    public const DELETE_NOTIFICATIONS_ERROR = '0';
    use LiberoWachtlTrait;
    use TranslatorAwareTrait;

    /**
     * Contains after init() is done all the limitations that are configured in the
     * FincLibero.ini as being used for triggering actions etc (e.g.:
     * requestableLimitations).
     *
     * @var array
     */
    protected $configuredLimitations = [];

    /**
     * Limitations that will trigger the placeStorageRetrievalRequest action
     *
     * @var array
     */
    protected $requestableLimitations = [];

    /**
     * Limitations that will trigger the placeHold action
     *
     * @var array
     */
    protected $holdableLimitations = [];

    /**
     * Limitations that will trigger a recall action (currently not used).
     *
     * @var array
     */
    protected $recallableLimitations = [];

    /**
     * Limitations that will identify the item for being bound to another item.
     *
     * @var array
     */
    protected $awlLimitations = [];

    /**
     * Patterns that will identify limitations as pickUpLocations
     *
     * @var array
     */
    protected $pickUpLocationPatterns = [];

    /**
     * URIs that will be used for stack views
     *
     * @var array
     */
    protected $stackURIs = [];

    /**
     * URIs that will be used for reading room views
     *
     * @var array
     */
    protected $readingRoomURIs = [];

    /**
     * Regex pattern to detect bound item ID
     * @var string
     */
    protected $boundItemIdPattern;

    /**
     * Regex pattern to detect bound item label
     * @var string
     */
    protected $boundItemLabelPattern;

    public function init()
    {
        parent::init();

        $this->setMemberFromConfig('boundItemIdPattern', 'General', 'bound_item_id_pattern');
        $this->setMemberFromConfig(
            'boundItemLabelPattern',
            'General',
            'bound_item_label_pattern'
        );

        // Get the base URI to extend departments.
        $this->setMemberFromConfig(
            'departmentLocationBase',
            'General',
            'departmentLocationBase',
            "No departmentLocationBase defined in FincLibero.ini."
        );

        // get the URIs for the limitations that are marking an item that a hold can
        // be placed on
        $this->setMemberFromConfig(
            'titleHoldLimitations',
            'General',
            'titleHoldLimitations',
            "No Limitations defined for action titleHold.",
            'titleHoldLimitations'
        );

        // get the pickUpLocationPatterns configured in ILS ini and being used for
        // filtering limitations used for pickuplocations
        $this->setMemberFromConfig(
            'pickUpLocationPatterns',
            'General',
            'pickUpLocationPatterns',
            "No pickUpLocationPatterns defined in FincLibero.ini."
        );

        // get the URIs for the limitations that are marking an item that a storage
        // retrieval request can be placed on
        $this->setMemberFromConfig(
            'requestableLimitations',
            'General',
            'requestableLimitations',
            "No Limitations defined for action Storage Retrieval Request.",
            'requestableLimitations'
        );

        // get the URIs for the limitations that are marking an item that a hold can
        // be placed on
        $this->setMemberFromConfig(
            'holdableLimitations',
            'General',
            'holdableLimitations',
            "No Limitations defined for action Hold.",
            'holdableLimitations'
        );

        // get the URIs for the limitations that are marking an item that a recall can
        // be placed on
        $this->setMemberFromConfig(
            'recallableLimitations',
            'General',
            'recallableLimitations',
            "No Limitations defined for action Recall.",
            'recallableLimitations'
        );

        // get URIs of limitations that will identify the item for being bound to
        // another item
        $this->setMemberFromConfig(
            'awlLimitations',
            'General',
            'awlLimitations',
            "No awlLimitations defined.",
            'awlLimitations'
        );

        // get URIs of limitations that will be ignored by VuFind
        $this->setMemberFromConfig(
            'ignoriedLimitations',
            'General',
            'ignoriedLimitations',
            "No ignoriedLimitations defined.",
            'ignoriedLimitations'
        );

        // get the URIs identifying records for stack views
        $this->setMemberFromConfig(
            'stackURIs',
            'General',
            'stackURIs',
            "No stack URIs defined."
        );

        // get the URIs identifying records for reading room views
        $this->setMemberFromConfig(
            'readingRoomURIs',
            'General',
            'readingRoomURIs',
            "No reading room URIs defined."
        );
    }

    /**
     * Helper function for @see FincLibero::init()
     * Set some member variables based on Config entries from FincILS.ini or
     * FincLibero.ini.
     * If the given config key (in given section) is set, the member variable will be
     * initialized with the value(s) from that key.
     * Otherwise (not set):
     * When the setting is required via the [RequiredConfig] section AND a debug message is
     * given, this message will be added to the debug log.
     * With the last parameter $configuredLimitationName, the config entry will be added
     * to the set of configured limitations under the given name
     *
     * @param string $memberName               name of class member to be set
     * @param string $configSection            section from INI file to look up (e.g. 'General' for section [General])
     * @param string $configKey                config key within that section
     * @param string $debugMessage             message to add to debug log on fail if the config is set required
     * @param string $configuredLimitationName name of limitation type to be used in setConfiguredLimitation()
     */
    protected function setMemberFromConfig(
        $memberName,
        $configSection,
        $configKey,
        $debugMessage = '',
        $configuredLimitationName = ''
    ) {
        if (
            !isset($this->config[$configSection])
            ||
            !isset($this->config[$configSection][$configKey])
        ) {
            if (
                !empty($debugMessage)
                &&
                ($this->config['RequiredConfig'][$configSection][$configKey] ?? false)
            ) {
                throw new InitException($debugMessage);
            }
        } else {
            $this->{$memberName} = $this->config[$configSection][$configKey];
            if (!empty($configuredLimitationName)) {
                $this->setConfiguredLimitation($configuredLimitationName);
            }
        }
    }

    /**
     * Helper function to extract the Namespace from the DAIA URI prefix.
     *
     * @return string
     */
    public function getDaiaIdPrefixNamespace()
    {
        return preg_quote(substr($this->daiaIdPrefix, 0, strpos($this->daiaIdPrefix, ':') + 1));
    }

    /**
     * Parse an array with DAIA status information.
     * Copies \VuFind\ILS\Driver\DAIA::parseDaiaArray to enable finc-specific
     * titleHoldLogic
     *
     * @param string $id        Record id for the DAIA array.
     * @param array  $daiaArray Array with raw DAIA status information.
     *
     * @return array            Array with VuFind compatible status information.
     */
    protected function parseDaiaArray($id, $daiaArray)
    {
        $doc_id = null;
        $doc_href = null;

        if (isset($daiaArray['id'])) {
            $doc_id = $daiaArray['id'];
        }

        if (isset($daiaArray['href'])) {
            // url of the document (not needed for VuFind)
            $doc_href = $daiaArray['href'];
        }

        if (isset($daiaArray['message'])) {
            // log messages for debugging
            $this->logMessages($daiaArray['message'], 'document');
        }

        // if one or more items exist, iterate and build result-item
        if (isset($daiaArray['item']) && is_array($daiaArray['item'])) {
            $isTitleHold = null;
            $isTitleHoldable = $this->hasTitleHolds();
            $number = 0;
            foreach ($daiaArray['item'] as $item) {
                // if it is a title-holdable record, the first item is a dummy
                // that delivers the title hold information and MUST NOT be
                // forwarded as an actual piece of availability information
                // instead, all following items MUST be marked as title-holdable,
                // to enable the frontend to display an adequate link
                if ($isTitleHoldable && !$isTitleHold) {
                    if ($isTitleHold = $this->isTitleHold($item)) {
                        $titleHoldId = $item['id'];
                        continue;
                    }
                }
                $result_item = [];
                $result_item['id'] = $id;

                // custom DAIA field
                $result_item['doc_id'] = $doc_id;
                if (
                    $isTitleHold
                    &&
                    (
                        !isset($this->noTitleHoldStatuses)
                    || empty($this->noTitleHoldStatuses)
                    || !in_array($item['localIlsStatus'], $this->noTitleHoldStatuses)
                    )
                ) {
                    $result_item['item_id'] = $titleHoldId;
                } else {
                    $result_item['item_id'] = $item['id'];
                }

                // custom DAIA field used in getHoldLink()
                $result_item['ilslink']
                    = ($item['href'] ?? $doc_href);
                // show TitleHoldLink except for defined noTitleHoldStatuses
                if ($isTitleHold && !in_array($item['localIlsStatus'], $this->noTitleHoldStatuses)) {
                    $result_item['addTitleHoldLink'] = true;
                }

                // count items
                $number++;
                $result_item['number'] = $this->getItemNumber($item, $number);

                // set default value for barcode
                $result_item['barcode'] = $this->getItemBarcode($item);

                // set default value for reserve
                $result_item['reserve'] = $this->getItemReserveStatus($item);

                // get callnumber
                $result_item['callnumber'] = $this->getItemCallnumber($item);

                // get location
                $result_item['location'] = $this->getItemDepartment($item);

                // custom DAIA field
                $result_item['locationid'] = $this->getItemDepartmentId($item);

                // get location link
                $result_item['locationhref'] = $this->getItemDepartmentLink($item);

                // custom DAIA field
                $result_item['storage'] = $this->getItemStorage($item);

                // custom DAIA field
                $result_item['storageid'] = $this->getItemStorageId($item);

                // custom DAIA field
                $result_item['storagehref'] = $this->getItemStorageLink($item);

                // status and availability will be calculated in own function
                $result_item = $this->getItemStatus($item) + $result_item;

                // add result_item to the result array
                $result[] = $result_item;
            } // end iteration on item
        }
        return $result;
    }

    /**
     * FincLibero specific overrides of PAIA methods
     */

    /**
     * Place Hold
     *
     * Attempts to place a hold or recall on a particular item and returns
     * an array with result details
     *
     * Make a request on a specific record
     *
     * @param array $holdDetails An array of item and patron data
     *
     * @return mixed An array of data on the request including
     * whether or not it was successful and a system message (if available)
     */
    public function placeHold($holdDetails)
    {
        $item = $holdDetails['item_id'];
        $patron = $holdDetails['patron'];

        $doc = [];
        $doc['item'] = stripslashes($item);
        if ($confirm = $this->getConfirmations($holdDetails)) {
            $doc["confirm"] = $confirm;
        }
        $post_data['doc'][] = $doc;

        try {
            $array_response = $this->paiaPostAsArray(
                'core/' . $patron['cat_username'] . '/request',
                $post_data
            );
        } catch (Exception $e) {
            $this->debug($e->getMessage());
            return [
                'success' => false,
                'sysMessage' => $e->getMessage(),
            ];
        }

        $details = [];

        if (array_key_exists('error', $array_response)) {
            $details = [
                'success' => false,
                'sysMessage' => $array_response['error_description']
            ];
        } else {
            $elements = $array_response['doc'];
            foreach ($elements as $element) {
                if (array_key_exists('error', $element)) {
                    $details = [
                        'success' => false,
                        'sysMessage' => $element['error']
                    ];
                } else {
                    // FincLibero supports more expressive responses from paialibero
                    // which should be shown instead to the user (localIlsStatus)
                    $details = [
                        'success' => true,
                        'sysMessage' => $element['localIlsStatus'] ?? 'Successfully requested'
                    ];
                    // if caching is enabled for DAIA remove the cached data for the
                    // current item otherwise the changed status will not be shown
                    // before the cache expires
                    if ($this->daiaCacheEnabled) {
                        $this->removeCachedData($holdDetails['doc_id']);
                    }
                    if ($this->paiaCacheEnabled) {
                        $this->removeCachedData($patron['cat_username']);
                    }
                }
            }
        }
        return $details;
    }

    /**
     * PAIA authentication function with custom signature -- libraryuserid added for
     * Libero Id
     *
     * @param string $username      Username
     * @param string $password      Password
     * @param string $libraryuserid Libero ID
     *
     * @return mixed Associative array of patron info on successful login,
     * null on unsuccessful login, PEAR_Error on error.
     * @throws ILSException
     */
    protected function paiaLogin($username, $password, $libraryuserid = null)
    {
        // perform full PAIA auth and get patron info
        $post_data = [
            "username"   => $username,
            "password"   => $password,
            "grant_type" => "password",
            "scope"      => self::SCOPE_READ_PATRON . " " .
                self::SCOPE_READ_FEES . " " .
                self::SCOPE_READ_ITEMS . " " .
                self::SCOPE_WRITE_ITEMS . " " .
                self::SCOPE_CHANGE_PASSWORD
        ];

        // Customization for internal Libero Id
        if (isset($libraryuserid)) {
            $post_data['libraryuserid'] = $libraryuserid;
        }

        $responseJson = $this->paiaPostRequest('auth/login', $post_data);

        try {
            $responseArray = $this->paiaParseJsonAsArray($responseJson);
        } catch (Exception $e) {
            // all error handling is done in paiaHandleErrors so pass on the excpetion
            throw $e;
        }

        if (!isset($responseArray['access_token'])) {
            throw new ILSException(
                'Unknown error! Access denied.'
            );
        } elseif (!isset($responseArray['patron'])) {
            throw new ILSException(
                'Login credentials accepted, but got no patron ID?!?'
            );
        } else {
            // at least access_token and patron got returned which is sufficient for
            // us, now save all to session
            $session = $this->getSession();

            // as we do not store the password in the database, write it to the
            // session (but only if we are not using root-login
            if ($libraryuserid === null) {
                $session->cat_password = $password;
            }

            $session->patron
                = $responseArray['patron'] ?? null;
            $session->access_token
                = $responseArray['access_token'] ?? null;
            $session->scope
                = isset($responseArray['scope'])
                ? explode(' ', $responseArray['scope']) : null;
            $session->expires
                = isset($responseArray['expires_in'])
                ? (time() + ($responseArray['expires_in'])) : null;
            // Customization for internal Libero Id
            $session->borrower_id
                = $responseArray['borrower_id'] ?? null;

            return true;
        }
    }

    /**
     * FincLibero specific override: we need to manually update the cat_password in
     * the session as we do not save passwords in the database and therefore the
     * ILSAuthenticator will use the session data instead.
     *
     * @param array $details Array with patron information, newPassword and
     *                       oldPassword.
     *
     * @return array An array with patron information.
     */
    public function changePassword($details)
    {
        $retval = parent::changePassword($details);

        if ($retval == ['success' => true, 'status' => 'Successfully changed']) {
            $session = $this->getSession();
            $session->cat_password = $details['newPassword'];
        }

        return $retval;
    }

    /**
     * PAIA helper function to map session data to return value of patronLogin()
     *
     * @param $details  Patron details returned by patronLogin
     * @param $password Patron cataloge password
     * @return mixed
     */
    protected function enrichUserDetails($details, $password, $username = null)
    {
        $details = parent::enrichUserDetails($details, $password, $username);

        // store also the Libero Id
        $session = $this->getSession();
        $details['borrower_id']  = $session->borrower_id;
        return $this->enrichWithUserRestrictions($details);
    }

    /**
     * Determines for each of the tokens in @param $details['type'] whether it is
     * configured to be a restriction type (as configured in FincLibero.ini
     * [General] userRestrictionMessagePatterns) and saves positive matches in
     * $details['restrictions'].
     *
     * @param $details patron details
     * @return mixed modified patron details
     */
    protected function enrichWithUserRestrictions($details)
    {
        //early return
        if (!isset($details['type'])
            || empty($details['type'])
            || !isset($this->config['General']['userRestrictionMessagePatterns'])
        ) {
            return $details;
        }

        foreach ($this->config['General']['userRestrictionMessagePatterns'] as $pattern) {
            foreach ($details['type'] as $type) {
                if (preg_match($pattern, $type)) {
                    $details['restrictions'][] = 'UserRestrictionMessages::' . $type;
                }
            }
        }
        return $details;
    }

    /**
     * Returns the value for "callnumber" in VuFind getStatus/getHolding array
     *
     * @param array $item Array with DAIA item data
     *
     * @return string
     */
    protected function getItemCallnumber($item)
    {
        return isset($item['label']) && !empty($item['label'])
            ? $item['label']
            : '';
    }

    /**
     * FincLibero specific overrides of DAIA methods
     */

    /**
     * Returns an array with status information for provided item.
     *
     * @param array $item Array with DAIA item data
     *
     * @return array
     */
    protected function getItemStatus($item)
    {
        $return = parent::getItemStatus($item);

        $return['item_notes'] = $this->getItemNotes($return, $item);

        $return['awlRecordId'] = $this->getBoundItemId($item);
        // is this item bound with another item?
        if ($return['awlRecordId'] != null) {
            static $awlStatuses = [];
            // overwrite any existing link settings as we need to order this item
            // via the bound item
            $return['addLink'] =
            $return['addStorageRetrievalRequestLink'] =
            $return['addILLRequestLink'] =
            $return['addEmailHoldLink'] = false;
            if (isset($awlStatuses[$return['awlRecordId']])) {
                $return['awlRecordStatus'] = $awlStatuses[$return['awlRecordId']];
            } else {
                // fix for infinite recursion due to circularly bound items
                $awlStatuses[$return['awlRecordId']] = 'pending';
                $return['awlRecordStatus'] =
                $awlStatuses[$return['awlRecordId']] =
                    current($this->getStatus($return['awlRecordId']));
            }
        }

        $return['service_type'] = $this->_reduceServices($return['services']);

        return $return;
    }

    /**
     * Helper function for getItemStatus().
     * Gather all item notes from current (parent) result and DAIA item
     *
     * @param array $return intermediate result from parent::getItemStatus()
     * @param array $item   DAIA item
     * @return array 'item_notes' part for result
     */
    protected function getItemNotes($return, $item)
    {
        // add all item specific information from DAIA field about to item_notes
        // (https://intern.finc.info/issues/7863)
        $about = $item['about'] ?? [];
        return array_unique(
            array_merge(
                (array)$return['status'],
                $return['item_notes'],
                (array)$about
            )
        );
    }

    /**
     * Helper function for 'service_type' entry in getItemStatus
     *
     * @param array $services actually present services in DAIA item
     * @return false|int|mixed preferred service if present or 0
     */
    protected function _reduceServices($services)
    {
        $prio = [
            'loan',
            'presentation',
        ];
        $res = array_intersect($prio, $services);
        if (empty($res)) {
            return 0;
        } else {
            return current($res);
        }
    }

    /**
     * Helper function to return an appropriate status string for current item
     *
     * @param $item
     * @return string
     */
    protected function getStatusString($item)
    {
        return $item['localIlsStatus'] ?? '';
    }

    /**
     * FincLibero specific implementation that adds pickUpLocations and the content
     * of EmailHold limitations to the customData array.
     *
     * @param array $item A DAIA item.
     * @return array
     */
    protected function getCustomData($item)
    {
        $customData = [];
        $customData['pickUpLocations'] = [];
        // flag to indicate that we have a valid service for loan or presentation
        $hasValidService = false;
        foreach (['available', 'unavailable'] as $availability) {
            if (isset($item[$availability])) {
                foreach ($item[$availability] as $available) {
                    if (isset($available['service'])
                        && in_array($available['service'], ['presentation', 'loan'])
                    ) {
                        $hasValidService = true;
                        // deal with pickuplocations
                        if (isset($available['limitation'])
                            && $pickUpLocations = $this->filterPickUpLocations(
                                $available['limitation']
                            )
                        ) {
                            // if we have limitations qualifying for pickUpLocations,
                            // save those in customData
                            $customData['pickUpLocations'] = array_merge(
                                $customData['pickUpLocations'],
                                $pickUpLocations
                            );
                        }
                        // deal with EmailHold information
                        if (isset($available['limitation'])) {
                            foreach ($available['limitation'] as $limitation) {
                                $criteria = $this->getEmailHoldValidationCriteria();
                                // we assume that each configured criteria for
                                // EmailHold will provide information (if existent in
                                // current item) supposed to be passed to the view
                                // via customData array
                                foreach ($criteria as $key => $value) {
                                    if ($this->checkEmailHoldValidationCriteria(
                                        [$key=>$limitation['id']]
                                    )
                                    ) {
                                        $customData['emailHoldLimitationContent'][] = $limitation['content'];
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        // check if we have a valid service, but no pickup locations given, then set fallback pickup location department
        if ($hasValidService && empty($customData['pickUpLocations']) && isset($item['department'])) {
            // if we have no explicit limitations qualifying for
            // pickUpLocations, assume the item's department as single
            // pickUpLocation
            $customData['pickUpLocations'] = [
                [
                    'locationID' => $item['department']['id'],
                    'locationDisplay' => $item['department']['content']
                ]
            ];
        }
        return $customData;
    }

    /**
     * Returns the value of item.storage.content (e.g. to be used in VuFind
     * getStatus/getHolding array as location)
     *
     * @param array $item Array with DAIA item data
     *
     * @return string
     */
    protected function getItemDepartment($item)
    {
        return $this->translate(parent::getItemDepartment($item));
    }

    /**
     * Returns the evaluated values of the provided limitations element -- FincLibero
     * specific implementation preprocesses limitations by filtering the non
     * functional limitations (i.e. limitations that are used for triggering actions
     * etc.)
     *
     * @param array $limitations Array with DAIA limitation data
     *
     * @return array
     */
    protected function getItemLimitationContent($limitations)
    {
        return parent::getItemLimitationContent(
            $this->filterNonFunctionalLimitations($limitations)
        );
    }

    /**
     * Returns the evaluated values of the provided limitations element -- FincLibero
     * specific implementation preprocesses limitations by filtering the non
     * functional limitations (i.e. limitations that are used for triggering actions
     * etc.)
     *
     * @param array $limitations Array with DAIA limitation data
     *
     * @return array
     */
    protected function getItemLimitationTypes($limitations)
    {
        return parent::getItemLimitationTypes(
            $this->filterNonFunctionalLimitations($limitations)
        );
    }

    /**
     * FincLibero specific MyReSearch views
     */

    /**
     * Custom FincLibero-method to return items with properties:
     *      - document.status = 4
     *      - document.storage und document.storageid != Lesesaal/Magazin
     *
     * @param array $patron Array returned from patronLogin()
     *
     * @return array
     */
    public function getMyMediaReadyToPickup($patron)
    {
        // filters for getMyMediaReadyToPickup are:
        // status = 4 - provided (the document is ready to be used by the patron)
        // document.storage und document.storageid != Lesesaal/Magazin
        $filter = [
            'status'  => [4],
            /*'exclude' => [
                'storageid' => $this->stackURIs
            ],*/
            'endtime' => null,
            'regex'   => ['item' => "/^(" . $this->getDaiaIdPrefixNamespace() . ").*$/"]
        ];
        // get items-docs for given filters
        $items = $this->paiaGetItems($patron, $filter);

        return $this->mapPaiaItems($items, 'myHoldsMapping');
    }

    /**
     * Custom FincLibero-method to return items with properties:
     *      - document.status = 3
     *      - document.endtime = 0
     *
     * @param array $patron Array returned from patronLogin()
     *
     * @return array
     */
    public function getMyPermanentLoans($patron)
    {
        // filters for getMyPermanentLoans are:
        // status = 3 - held (the document is on loan by the patron)
        // endtime = 0
        $filter = [
            'status'  => [3],
            'endtime' => null,
            'regex'   => ['item' => "/^(" . $this->getDaiaIdPrefixNamespace() . ").*$/"]
        ];
        // get items-docs for given filters
        $items = $this->paiaGetItems($patron, $filter);

        return $this->mapPaiaItems(
            $this->itemFieldSort($items, 'starttime', SORT_DESC),
            'myTransactionsMapping'
        );
    }

    /**
     * Customized getMyHolds for FincLibero to return items with properties:
     *      - document.status = 1 (reserved)
     *      - document.storage und document.storageid != Magazin
     *      - document.endtime = Rückgabedatum (not NULL!)
     *
     * @param array $patron Array returned from patronLogin()
     *
     * @return array
     */
    public function getMyHolds($patron)
    {
        // filters for getMyHolds are:
        // status = 1 - reserved
        $filter = [
            'status'  => [1,2],
            'regex'   => ['item' => "/^(" . $this->getDaiaIdPrefixNamespace() . ").*$/"]
        ];
        // get items-docs for given filters
        $items = $this->paiaGetItems($patron, $filter);

        return $this->mapPaiaItems($items, 'myHoldsMapping');
    }

    /**
     * This method queries the ILS for a patron's current storage retrieval requests.
     * Returns items with properties:
     *      - document.status = 2 (ordered)
     *      - document.storage and document.storageid do have a certain location (Magazin, but not Lesesaal)
     *
     * note: stackURIs musst be configured in FincLibero.ini as required
     *
     * @param array $patron Array returned from patronLogin()
     *
     * @return array
     */
    public function getMyStorageRetrievalRequests($patron)
    {
        $filter = [
            'status'  => [2],
            'storageid' => $this->stackURIs,
            'regex'   => ['item' => "/^(" . $this->getDaiaIdPrefixNamespace() . ").*$/"]
        ];
        // get items-docs for given filters
        $items = $this->paiaGetItems($patron, $filter);

        return $this->mapPaiaItems($items, 'myStorageRetrievalRequestsMapping');
    }

    /**
     * Customized getMyTransactions for FincLibero to return items with properties:
     *      - document.status = 3 (held)
     *      - document.item = UBL:*
     *      - document.starttime = Ausleihdatum
     *      - document.endtime = Rückgabedatum (not NULL!)
     *
     * @param array $patron Array returned from patronLogin()
     *
     * @return array
     */
    public function getMyTransactions($patron)
    {
        // filters for getMyPermanentLoans are:
        // status = 3 - held (the document is on loan by the patron)
        // endtime != null
        $filter = [
            'status'  => [3],
            'exclude' => ['endtime' => null],
            'regex'   => ['item' => "/^(" . $this->getDaiaIdPrefixNamespace() . ").*$/"]
        ];
        // get items-docs for given filters
        $items = $this->paiaGetItems($patron, $filter);

        return $this->mapPaiaItems(
            $this->itemFieldSort($items, 'endtime'),
            'myTransactionsMapping'
        );
    }

    /**
     * @param $patron
     * @param null $messageIdList
     * @param null $toDate won't work here
     * @throws ILSException
     * @return TRUE on success, FALSE otherwise
     */
    public function removeMySystemMessages(
        $patron,
        $messageIdList = null,
        $toDate = null
    ) {
        return $this->paiaRemoveSystemMessages($patron, $messageIdList);
    }

    /**
     * FincLibero specific helper functions
     */

    /**
     * Helper function to filter certain limitations.
     *
     * @param array $limitations An item's limitations.
     * @return mixed
     */
    protected function filterNonFunctionalLimitations($limitations)
    {
        // remove the configured limitations from the current set of limitations
        foreach ($this->configuredLimitations as $configuredLimitation) {
            foreach ($limitations as $key => $limitation) {
                if (isset($limitation['id'])
                    && in_array($limitation['id'], $this->{$configuredLimitation})
                ) {
                    unset($limitations[$key]);
                }
            }
        }

        // remove all known pickUpLocations from the current set of limitations
        $pickUpLocations = $this->filterPickUpLocations($limitations);
        foreach ($limitations as $key => $limitation) {
            if (isset($limitation['id'])) {
                foreach ($pickUpLocations as $pickup) {
                    if ($limitation['id'] == $pickup['locationID']) {
                        unset($limitations[$key]);
                    }
                }
            }
        }

        return $limitations;
    }

    /**
     * Helper function to compile upon init() a set of limitations that can be used
     * for filtering.
     *
     * @param string $limitation An URI identifying a limitation
     */
    protected function setConfiguredLimitation($limitation)
    {
        if (is_string($limitation)) {
            $this->configuredLimitations[] = $limitation;
        }
    }

    /**
     * Helper function to filter PickUpLocations from given limitations based on
     * pattern configured in ILS ini
     *
     * @param array $limitations An item's limitations
     * @return array
     */
    protected function filterPickUpLocations($limitations)
    {
        $pickUpLocations = [];
        // return array(locationID=>URI, locationDisplay=>Content)
        foreach ($limitations as $limitation) {
            if (isset($limitation['id'])) {
                foreach ($this->pickUpLocationPatterns as $pattern) {
                    if (preg_match($pattern, $limitation['id'])
                        && !(filter_var($limitation['id'], FILTER_VALIDATE_URL) === false)
                    ) {
                        $pickUpLocations[] = [
                            'locationID' => $limitation['id'],
                            'locationDisplay' => $limitation['content'] ?? $limitation['id'],
                        ];
                    }
                }
            }
        }
        return $pickUpLocations;
    }

    /**
     * Helper method to check whether this item is bound with another item and needs
     * to be ordered via this bound item
     *
     * @param array $item Array with DAIA item data
     *
     * @return null
     */
    protected function getBoundItemId($item)
    {
        if (!isset($this->boundItemIdPattern)) {
            return null;
        }
        $availabilities = ['available', 'unavailable'];

        // start logic to check if we have a bound item that needs to be ordered
        // via another record (see https://intern.finc.info/issues/5068)
        foreach ($availabilities as $availability) {
            if (isset($item[$availability])) {
                foreach ($item[$availability] as $available) {
                    if (isset($available['limitation'])) {
                        foreach ($available['limitation'] as $limitation) {
                            if (in_array($limitation['id'], $this->awlLimitations)) {
                                // items with limitations identifying the record for
                                // being bound to another item contain the RSN of the
                                // bound item in their current available service href
                                if (isset($available['href'])
                                    && preg_match(
                                        $this->boundItemIdPattern,
                                        $available['href'],
                                        $matches
                                    )
                                ) {
                                    return $this->queryBoundItem($matches['id']);
                                }
                            }
                        }
                    }
                }
            }
        }

        // if we made it this far this item definitely is not bound to another item
        return null;
    }

    /**
     * Helper method to query bound item by rsn for record_id
     *
     * @param string $key
     *
     * @return string record_id
     */
    protected function queryBoundItem($key)
    {
        $query = $this->getBoundItemQueryString($key);
        if (empty($query)) {
            return null;
        }
        try {
            // let's search with the built query
            $result = $this->searchService
                ->search('VuFind', new Query($query));
            if (count($result) === 0) {
                $this->debug('Problem retrieving rsn ' .
                    'for record with query:' . $query);
                return null;
            }
            // pass the id from the first found record as
            // awlRecordId to the controller
            return current($result->getRecords())
                ->getUniqueId();
        } catch (\Exception $e) {
            $this->debug($e);
            return null;
        }
    }

    /**
     * Check if item is a pretended dummy to signify there is a title order
     * possible.
     *
     * @param array $item DAIA item
     *
     * @return mixed
     */
    public function isTitleHold($item)
    {
        foreach ($this->titleHoldLimitations as $limitation) {
            $regex = '/^' . trim($limitation) . '$/';
            if (0 < preg_match($regex, $item['id'])) {
                return true;
            }
        }
        return null;
    }

    /**
     * May there be Title holdable items in this instance?
     * Inherited FincLibero drivers should set see $titleHoldLimitations
     * to achieve this
     *
     * @return bool
     */
    public function hasTitleHolds()
    {
        return isset($this->titleHoldLimitations);
    }

    /**
     * Place Title Hold
     *
     * Attempts to place a hold or recall on a particular item and returns
     * an array with result details
     *
     * Make a request on a specific record
     *
     * @param array $holdDetails An array of item and patron data
     *
     * @return mixed An array of data on the request including
     * whether or not it was successful and a system message (if available)
     */
    public function placeTitleHold($holdDetails)
    {
        $item = $holdDetails['item_id'];
        $patron = $holdDetails['patron'];
        $doc = [];
        $doc['item'] = stripslashes($item);
        if ($confirm = $this->getConfirmations(
            $this->extendPickUpLocation($holdDetails)
        )) {
            $doc["confirm"] = $confirm;
        }
        $post_data['doc'][] = $doc;
        try {
            $array_response = $this->paiaPostAsArray(
                'core/' . $patron['cat_username'] . '/request',
                $post_data
            );
        } catch (Exception $e) {
            $this->debug($e->getMessage());
            return [
                'success' => false,
                'sysMessage' => $e->getMessage(),
            ];
        }
        $details = [];
        if (array_key_exists('error', $array_response)) {
            $details = [
                'success' => false,
                'sysMessage' => $array_response['error_description']
            ];
        } else {
            $elements = $array_response['doc'];
            foreach ($elements as $element) {
                if (array_key_exists('error', $element)) {
                    $details = [
                        'success' => false,
                        'sysMessage' => $element['error']
                    ];
                } else {
                    // FincLibero supports more expressive responses from paialibero
                    // which should be shown instead to the user (localIlsStatus)
                    $details = [
                        'success' => true,
                        'sysMessage' => $element['localIlsStatus'] ?? 'Successfully requested'
                    ];
                    // if caching is enabled for DAIA remove the cached data for the
                    // current item otherwise the changed status will not be shown
                    // before the cache expires
                    if ($this->daiaCacheEnabled) {
                        $this->removeCachedData($holdDetails['doc_id']);
                    }
                }
            }
        }
        return $details;
    }

    /**
     * Extend pickup location with base department URI
     *
     * @param array $holdDetails An array of item and patron data.
     *
     * @return array $holdDetails;
     */
    protected function extendPickUpLocation($holdDetails)
    {
        if (isset($holdDetails['pickUpLocation'])) {
            $holdDetails['pickUpLocation']
                = $this->departmentLocationBase . $holdDetails['pickUpLocation'];
        }
        return $holdDetails;
    }

    /**
     * Helper function for queryBoundItem. Prepares instance specific query string
     * to get the ID of the bound item
     *
     * Override in inherited FincLibero
     *
     * @param $key string   ID to search for
     * @return string       SOLR query to be used as q parameter
     */
    protected function getBoundItemQueryString($key)
    {
        return '';
    }

    /**
     * This method returns a list of locations where a user may collect a hold.
     *
     * @param array $patron  Patron array returned by patronLogin method
     * @param array $details Hold information array similar to placeHold's input
     * @return array Array of associative arrays containing these keys:
     *      locationID - A pick up location id or code (string)
     *      locationDisplay – The text to display for the location (string)
     *
     * @link https://vufind.org/wiki/development:plugins:ils_drivers#getpickuplocations
     */
    public function getPickUpLocations($patron = [], $details = [])
    {
        // use departments of Wachtl, not DAIA
        if ($this->config['General']['useWachtlPickupLocations'] ?? false) {
            return $this->getWachtlPickUpLocations($details);
        }

        // use location and locationid, not customData of DAIA
        if ($this->config['General']['useDaiaLocationId'] ?? false) {
            return parent::getPickUpLocations($patron, $details);
        }

        // use DAIA customData
        if ($details != [] && isset($details['id'])) {
            $statusItems = $this->getStatus($details['id']);
            foreach ($statusItems as $statusItem) {
                if ($statusItem['item_id'] == $details['item_id']) {
                    // our pickUpLocations are stored in the customData array upon
                    // processing the DAIA availability information
                    if (isset($statusItem['customData'])
                        && isset($statusItem['customData']['pickUpLocations'])
                    ) {
                        return $statusItem['customData']['pickUpLocations'];
                    } else {
                        return $this->stackURIs ?? [];
                    }
                }
            }
        }
        return [];
    }
}
