<?php
/**
 * ReDi Link Resolver Driver
 *
 * PHP version 5
 *
 * Copyright (C) Leipzig University Library 2015
 *
 * @category VuFind
 * @package  Resolver_Drivers
 * @author   Robert Lange <lange@ub.uni-leipzig.de>
 * @author   Gregor Gawol <gawol@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:plugins:link_resolver_drivers Wiki
 */
namespace finc\Resolver\Driver;

use VuFind\Resolver\Driver\Redi as RediBase;
use Zend\Dom\DOMXPath;

/**
 * Redi Link Resolver Driver
 *
 * @category VuFind
 * @package  Resolver_Drivers
 * @author   Robert Lange <lange@ub.uni-leipzig.de>
 * @author   Gregor Gawol <gawol@ub.uni-leipzig.de>
 * @license  http://opensource.org/licenses/gpl-2.0.php GNU General Public License
 * @link     https://vufind.org/wiki/development:plugins:link_resolver_drivers Wiki
 */
class Redi extends RediBase
{

    /**
     * Constructor
     *
     * @param string            $config     redi configuration
     * @param \Zend\Http\Client $httpClient HTTP client
     */
    public function __construct($config, \Zend\Http\Client $httpClient)
    {
        $this->baseUrl = $config->url;
        $this->httpClient = $httpClient;
    }

    /**
     * Fetch Links
     *
     * Fetches a set of links corresponding to an OpenURL
     *
     * @param string $openURL openURL (url-encoded)
     *
     * @return string         raw XML returned by resolver
     */
    public function fetchLinks($openURL)
    {
        $url = $this->getResolverUrl($openURL);
        $feed = $this->httpClient->setUri($url)->send()->getBody();
        return $feed;
    }

    /**
     * Parse the Redi XML response and return array with DOI information.
     *
     * @param DOMDocument $xml Loaded xml document
     *
     * @deprecated with VuFind 9 - cf. https://github.com/vufind-org/vufind/pull/2419
     * @return array Get back a array with title, URL and service_type
     */
    protected function parseDOI($xml)
    {
        $retval = [];

        $xpath = new DOMXPath($xml);

        $doiTerm = $xpath
            ->query("//dt[@class='doi_t']");
        $doiDefinition = $xpath
            ->query("//dd[@class='doi_d']");

        if ($doiTerm->length == $doiDefinition->length) {
            for ($i = 0; $i < $doiTerm->length; $i++) {
                $href = $xpath
                    ->query(".//@href", $doiDefinition->item($i))
                    ->item(0)->textContent;
                $retval[] = [
                    'title' => $doiTerm->item($i)->textContent
                        . $doiDefinition->item($i)->textContent,
                    'href' => $href,
                    'access' => 'unknown',
                    'coverage' => null,
                    'service_type' => 'getDOI',
                ];
            }
        }

        return $retval;
    }
}
